///////////////////////////////////////////////////////////////////////////////
/*! \file Lighting.h
 *  \brief Framework Nemo
 *  \author C. Crassin
 *
 * $Id: UtilsGL.h,v 1.2 2007/09/14 11:54:56 gascuel Exp $
 */
///////////////////////////////////////////////////////////////////////////////
// History:
//
// 2005-01-01 : Creation.

#ifndef NemoGraphics_UtilsGL_h
#define NemoGraphics_UtilsGL_h

#ifdef WIN32
#include <windows.h>
#endif


//#define GLEW_STATIC

///////////////////////
#include <GL/glew.h>
#ifdef WIN32
	#include <GL/wglew.h>
#endif


#include <GL/glu.h>

#include <time.h>
#include <math.h>

#include <iostream>


#include <NemoGraphics/Vector.h>
#include <NemoGraphics/Matrix.h>
#include <NemoGraphics/ResourcesManager.h>
#include <NemoGraphics/MessageHandler.h>


#define BUFFER_OFFSET(i) ((char *)NULL + (i))

#define INT_INDICE_BUG 32768

//65536
#ifndef PI
#define PI 3.1415926535897932384626433832795
#endif

#define myround(x) (x<0?ceilf((x)-0.5):floorf((x)+0.5))
#define myroundf(x) (x<0.0f?ceilf((x)-0.5f):floorf((x)+0.5f))

namespace NemoGraphics {


inline void ngVertex(Vector4f &v){
	glVertex4fv((GLfloat*)&v);
}

inline void ngColor4f(Vector4f &v){
	glColor4fv((GLfloat*)&v);
}

inline int clamp(int v, const int &v0, const int &v1){
	if(v<v0)
		v=v0;
	if(v>v1)
		v=v1;
	return v;
}

inline void GLErrorCheck(const char *msg=NULL){
//#ifdef _NEMO_GL_ERROR_CHECK
	GLuint e=glGetError();
	if(e){
		char *es;
		if(e==GL_INVALID_FRAMEBUFFER_OPERATION_EXT){
			es="GL_INVALID_FRAMEBUFFER_OPERATION_EXT";
		}else{
			es=(char*)gluErrorString(e);
		}

		NEMO_msg().critical("%s OpenGL Error (%d): %s", msg, (int)e, es);
	}
//#endif
}

inline unsigned long hashString(const char *str) {
    unsigned long hash = 0;
    int c;
    while (c = *str++)
        hash = c + (hash << 6) + (hash << 16) - hash;

    return hash;
}

//////////////////////////////

inline std::string operator +(const std::string &str, const float f){
	char buff[4096];

	sprintf(buff, "%s%f", str.c_str(), f);

	return buff;
}

inline std::string operator +(const std::string &str, const int f){
	char buff[4096];

	sprintf(buff, "%s%d", str.c_str(), f);

	return buff;
}

#ifndef NG_USE_NEW_VEC

//get format string
template<class T>
inline std::string GetTypeString(){
	return 0;
}

template<>
inline std::string GetTypeString< float >(){
	return "FLOAT";
}

template<>
inline std::string GetTypeString< Vector4f >(){
	return "Vector4f";
}
template<>
inline std::string GetTypeString<Vector4i>(){
	return "Vector4i";
}
template<>
inline std::string GetTypeString<Vector4uc>(){
	return "Vector4uc";
}

template<>
inline std::string GetTypeString<unsigned char>(){
	return "UC";
}

template<>
inline std::string GetTypeString<int>(){
	return "INT";
}

////////////////////////////////

inline int GLGetFormatSize(GLuint format){
	switch(format){
		case GL_RGBA8:
		case GL_RGBA:
		case GL_BGRA:
			return 4;
		case GL_RGB8:
		case GL_RGB:
			return 3;
		case GL_RGBA16F_ARB:
			return 8;
		case GL_RGB16F_ARB:
			return 6;
		case GL_RGBA32F_ARB:
			return 16;
		case GL_RGB32F_ARB:
			return 12;
		default:
			return -1;
	}
}

//get gl types from c++
template<class T>
inline int GLGetGLType(){
	return 0;
}

template<>
inline int GLGetGLType<GLhalf>(){
	return GL_HALF_FLOAT_NV;
}

template<>
inline int GLGetGLType<Vector4<GLhalf>>(){
	return GL_HALF_FLOAT_NV;
}

template<>
inline int GLGetGLType< Vector4f >(){
	return GL_FLOAT;
}
template<>
inline int GLGetGLType< Vector4uc >(){
	return GL_UNSIGNED_BYTE;
}

template<>
inline int GLGetGLType< Vector3f >(){
	return GL_FLOAT;
}
template<>
inline int GLGetGLType< float >(){
	return GL_FLOAT;
}
template<>
inline int GLGetGLType< unsigned char >(){
	return GL_UNSIGNED_BYTE;
}


//get GL formats
template<class T>
inline GLuint GLGetGLExtFormat(){
	return 0;
}

template<>
inline GLuint GLGetGLExtFormat< Vector4f >(){
	return GL_RGBA;
}
template<>
inline GLuint GLGetGLExtFormat<Vector4i>(){
	return GL_RGBA;
}
template<>
inline GLuint GLGetGLExtFormat<Vector4uc>(){
	return GL_BGRA;
}

template<>
inline GLuint GLGetGLExtFormat<Vector3f>(){
	return GL_RGB;
}

template<>
inline GLuint GLGetGLExtFormat<float>(){
	return GL_ALPHA;
}
template<>
inline GLuint GLGetGLExtFormat<unsigned char>(){
	return GL_ALPHA;
}

//get GL formats
template<class T>
inline GLuint GLGetGLInternalFormat(){
	return 0;
}

template<>
inline GLuint GLGetGLInternalFormat< Vector4f >(){
	return GL_RGBA32F_ARB;
}
template<>
inline GLuint GLGetGLInternalFormat<Vector3f>(){
	return GL_RGB32F_ARB;
}
template<>
inline GLuint GLGetGLInternalFormat<float>(){
	return GL_ALPHA32F_ARB; //GL_LUMINANCE32F_ARB
	
}

template<>
inline GLuint GLGetGLInternalFormat<Vector4uc>(){
	return GL_RGBA8;
}
template<>
inline GLuint GLGetGLInternalFormat<unsigned char>(){
	return GL_ALPHA8;
}

inline GLuint GLGetNaturalExternalFormat(GLuint iformat){

	switch(iformat){
		////Integer formats////
		case GL_RGBA32I_EXT:
		case GL_RGBA16I_EXT:
			return GL_RGBA_INTEGER_EXT;
		case GL_RGB32I_EXT:
		case GL_RGB16I_EXT:
			return GL_RGB_INTEGER_EXT;
		//////////////////////

		/*case GL_COMPRESSED_RGBA_S3TC_DXT5_EXT:
			return GL_ALPHA;*/

		case GL_RGBA32F_ARB:
		case GL_RGBA16F_ARB:
			return GL_RGBA;
		case GL_RGBA8:
			return GL_BGRA;

		case GL_RGB32F_ARB:
		case GL_RGB16F_ARB:
		case GL_RGB8:
			return GL_RGB;

		case GL_ALPHA32F_ARB:
		case GL_ALPHA16F_ARB:
		case GL_ALPHA8:
			return GL_ALPHA;

		default:
			return 0;
	}
}

inline GLuint GLGetNaturalType(GLuint iformat){

	switch(iformat){
		////Integer formats
		case GL_RGBA32I_EXT:
		case GL_RGBA16I_EXT:
			return GL_INT;
		case GL_RGB32I_EXT:
		case GL_RGB16I_EXT:
			return GL_INT;
		//////////////////////

		case GL_RGBA32F_ARB:
		case GL_RGBA16F_ARB:
		case GL_RGB32F_ARB:
		case GL_RGB16F_ARB:
		case GL_ALPHA32F_ARB:
		case GL_ALPHA16F_ARB:
			return GL_FLOAT;

		case GL_RGBA8:	
		case GL_RGB8:
		case GL_ALPHA8:
			return GL_UNSIGNED_BYTE;

		default:
			return 0;
	}
}

inline GLint GLGetTextureTargetBinding(GLint target){
	switch(target){
		////Integer formats
		case GL_TEXTURE_1D :
			GL_TEXTURE_BINDING_1D;
			break;
		case GL_TEXTURE_2D :
			GL_TEXTURE_BINDING_2D;
			break;
		case GL_TEXTURE_3D :
			GL_TEXTURE_BINDING_3D;
			break;

		case GL_TEXTURE_RECTANGLE_EXT :
			GL_TEXTURE_BINDING_RECTANGLE_EXT;
			break;

		case GL_TEXTURE_CUBE_MAP :
			GL_TEXTURE_BINDING_CUBE_MAP;
			break;

		case GL_TEXTURE_1D_ARRAY_EXT :
			GL_TEXTURE_BINDING_1D_ARRAY_EXT;
			break;
		case GL_TEXTURE_2D_ARRAY_EXT :
			GL_TEXTURE_BINDING_2D_ARRAY_EXT;
			break;

		default:
			return 0;
	}
}


/////////Data conversions
template<class TS, class TD>
inline TD DataConversion(const TS &s){
	return TD(s);
}

template< >
inline Vector4f DataConversion<unsigned char, Vector4f>(const unsigned char &s){
	return Vector4f(float(s)/255.0f);
}

template< >
inline Vector4uc DataConversion<unsigned char, Vector4uc>(const unsigned char &s){
	return Vector4uc(s);
}

template< >
inline float DataConversion<unsigned char, float>(const unsigned char &s){
	return float(s)/255.0f;
}


#endif

/////////Gestion des extensions
class GLExtensionsManager {
	static bool initialized;
public:
	static int initialize();
	static int checkSupported(const char *exts);
};

//////////////////////////////
inline void drawCubeTex(Vector3f &p1, Vector3f &p2) {

	glBegin(GL_QUADS);
		// Front Face
		glTexCoord2f(0.0f, 0.0f); glVertex3f(p1.x, p1.y,  p2.z);	// Bottom Left Of The Texture and Quad
		glTexCoord2f(0.0f, 1.0f); glVertex3f(p1.x,  p2.y,  p2.z);	// Top Left Of The Texture and Quad
		glTexCoord2f(1.0f, 1.0f); glVertex3f( p2.x,  p2.y,  p2.z);	// Top Right Of The Texture and Quad
		glTexCoord2f(1.0f, 0.0f); glVertex3f( p2.x, p1.y,  p2.z);	// Bottom Right Of The Texture and Quad
		
		
		// Back Face
		glTexCoord2f(1.0f, 0.0f); glVertex3f(p1.x, p1.y, p1.z);	// Bottom Right Of The Texture and Quad
		glTexCoord2f(0.0f, 0.0f); glVertex3f( p2.x, p1.y, p1.z);	// Bottom Left Of The Texture and Quad
		glTexCoord2f(0.0f, 1.0f); glVertex3f( p2.x,  p2.y, p1.z);	// Top Left Of The Texture and Quad
		glTexCoord2f(1.0f, 1.0f); glVertex3f(p1.x,  p2.y, p1.z);	// Top Right Of The Texture and Quad
		 


		// Top Face
		glTexCoord2f(0.0f, 1.0f); glVertex3f(p1.x,  p2.y, p1.z);	// Top Left Of The Texture and Quad
		glTexCoord2f(1.0f, 1.0f); glVertex3f( p2.x,  p2.y, p1.z);	// Top Right Of The Texture and Quad
		glTexCoord2f(1.0f, 0.0f); glVertex3f( p2.x,  p2.y,  p2.z);	// Bottom Right Of The Texture and Quad
		glTexCoord2f(0.0f, 0.0f); glVertex3f(p1.x,  p2.y,  p2.z);	// Bottom Left Of The Texture and Quad
		
		
		
		// Bottom Face
		glTexCoord2f(1.0f, 1.0f); glVertex3f(p1.x, p1.y, p1.z);	// Top Right Of The Texture and Quad
		glTexCoord2f(1.0f, 0.0f); glVertex3f(p1.x, p1.y,  p2.z);	// Bottom Right Of The Texture and Quad
		glTexCoord2f(0.0f, 0.0f); glVertex3f( p2.x, p1.y,  p2.z);	// Bottom Left Of The Texture and Quad
		glTexCoord2f(0.0f, 1.0f); glVertex3f( p2.x, p1.y, p1.z);	// Top Left Of The Texture and Quad
		
		
		
		// Right face
		glTexCoord2f(1.0f, 0.0f); glVertex3f( p2.x, p1.y, p1.z);	// Bottom Right Of The Texture and Quad
		glTexCoord2f(0.0f, 0.0f); glVertex3f( p2.x, p1.y,  p2.z);	// Bottom Left Of The Texture and Quad
		glTexCoord2f(0.0f, 1.0f); glVertex3f( p2.x,  p2.y,  p2.z);	// Top Left Of The Texture and Quad
		glTexCoord2f(1.0f, 1.0f); glVertex3f( p2.x,  p2.y, p1.z);	// Top Right Of The Texture and Quad
		
		// Left Face
		glTexCoord2f(0.0f, 0.0f); glVertex3f(p1.x, p1.y, p1.z);	// Bottom Left Of The Texture and Quad
		glTexCoord2f(0.0f, 1.0f); glVertex3f(p1.x,  p2.y, p1.z);	// Top Left Of The Texture and Quad
		glTexCoord2f(1.0f, 1.0f); glVertex3f(p1.x,  p2.y,  p2.z);	// Top Right Of The Texture and Quad
		glTexCoord2f(1.0f, 0.0f); glVertex3f(p1.x, p1.y,  p2.z);	// Bottom Right Of The Texture and Quad
		
	glEnd();


}

inline void drawCube(Vector3f &p1, Vector3f &p2) {

	glBegin(GL_QUADS);
		// Front Face
		glVertex3f(p1.x, p1.y,  p2.z);	// Bottom Left Of The Texture and Quad
		glVertex3f(p1.x,  p2.y,  p2.z);	// Top Left Of The Texture and Quad
		glVertex3f( p2.x,  p2.y,  p2.z);	// Top Right Of The Texture and Quad
		glVertex3f( p2.x, p1.y,  p2.z);	// Bottom Right Of The Texture and Quad
	
		// Back Face
		glVertex3f(p1.x, p1.y, p1.z);	// Bottom Right Of The Texture and Quad
		glVertex3f( p2.x, p1.y, p1.z);	// Bottom Left Of The Texture and Quad
		glVertex3f( p2.x,  p2.y, p1.z);	// Top Left Of The Texture and Quad
		glVertex3f(p1.x,  p2.y, p1.z);	// Top Right Of The Texture and Quad
		 
		// Top Face
		glVertex3f(p1.x,  p2.y, p1.z);	// Top Left Of The Texture and Quad
		glVertex3f( p2.x,  p2.y, p1.z);	// Top Right Of The Texture and Quad
		glVertex3f( p2.x,  p2.y,  p2.z);	// Bottom Right Of The Texture and Quad
		glVertex3f(p1.x,  p2.y,  p2.z);	// Bottom Left Of The Texture and Quad
		
		// Bottom Face
		glVertex3f(p1.x, p1.y, p1.z);	// Top Right Of The Texture and Quad
		glVertex3f(p1.x, p1.y,  p2.z);	// Bottom Right Of The Texture and Quad
		glVertex3f( p2.x, p1.y,  p2.z);	// Bottom Left Of The Texture and Quad
		glVertex3f( p2.x, p1.y, p1.z);	// Top Left Of The Texture and Quad
		
		// Right face
		glVertex3f( p2.x, p1.y, p1.z);	// Bottom Right Of The Texture and Quad
		glVertex3f( p2.x, p1.y,  p2.z);	// Bottom Left Of The Texture and Quad
		glVertex3f( p2.x,  p2.y,  p2.z);	// Top Left Of The Texture and Quad
		glVertex3f( p2.x,  p2.y, p1.z);	// Top Right Of The Texture and Quad
		
		// Left Face
		glVertex3f(p1.x, p1.y, p1.z);	// Bottom Left Of The Texture and Quad
		glVertex3f(p1.x,  p2.y, p1.z);	// Top Left Of The Texture and Quad
		glVertex3f(p1.x,  p2.y,  p2.z);	// Top Right Of The Texture and Quad
		glVertex3f(p1.x, p1.y,  p2.z);	// Bottom Right Of The Texture and Quad
		
	glEnd();


}

};

#endif
